//
// C++ Implementation: %{MODULE}
//
// Description:
//
//
// Author: %{AUTHOR} <%{EMAIL}>, (C) %{YEAR}
//
// Copyright: See COPYING file that comes with this distribution
//
//
#include "cinfosat.h"
#include "section.h"
#include "hexdump.h"
#include "util.h"

#include <netinet/in.h>
#include <stdio.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <fcntl.h>
#include <unistd.h>
#include <string>
#include <sstream>
#include <sys/mman.h>
#include <iostream>
#include <iomanip>
#include <fstream>

#include <time.h>

using namespace std;

#define TABLE_SHORT    0x00
#define TABLE_EPG      0x01
#define TABLE_FIRMWARE 0x1a

// #define my_debug
#define standalone

#ifdef standalone
#define infosat_printf printf
#else
#define infosat_printf //
#endif

#ifdef my_debug
#define dbg_printf printf
#else
#define dbg_printf //
#endif


static ofstream *ilog=NULL;

string int2string(int number, bool hex=false)
{
   static char dest[10];
   snprintf(dest,10,hex ? "%02x" : "%02d",number);
   return string(dest);
}


const int PersistentStatus::mapSize=1024*1024;
const char PersistentStatus::id[12]="infosatEPG ";


int daydata_BIN::mismatched=0;
time_t daydata_BIN::programStartTime=time(NULL);

void daydata_BIN::setPacketDone(int i)
{
	if (i>=packetsTotal) {
		cerr << "Wrong Packet passed through, this should never happen" << endl;
		return;
	}
	if (goodPackets[i]==0) {
		goodPacketCount++;
	}
	goodPackets[i]=1;
	local.data.bFinished=(goodPacketCount == packetsTotal);
}

string daydata_BIN::getfname(void)
{
	return string("infosat_") + int2string(month) + "_" + int2string(day);
}

void daydata_BIN::init(InfosatHdr *ishdr)
{
	goodPacketCount=0;
	packetsTotal = ishdr->getPktCount();
	day = ishdr->getDay();
	month = ishdr->getMonth();
	macID = ishdr->getMacID();
	for (int i=0;i<packetsTotal; i++) {
		goodPackets[i]=0;
	}
	wakeup(true);
}

void daydata_BIN::wakeup(bool truncate)
{
	local.data.bFinished=(packetsTotal==goodPacketCount);
	local.data.bLogged=false;
	local.data.fd=-1;
	local.data.truncate=truncate;
	local.data.dayoffset=calcDayOffset();
	log();
}

void daydata_BIN::close()
{
	if (local.data.fd!=-1) {
		::close(local.data.fd);
		if (!finished()) {
			unlink(getfname().c_str());
		}
	}
	local.data.fd=-1;
	log();
}

void daydata_BIN::log()
{
	if (local.data.bFinished && ilog && !local.data.bLogged) {
		*ilog << getfname() << endl;
		local.data.bLogged=true;
	}
}

void daydata_BIN::write(InfosatHdr *ishdr, RawData Data, int &offset)
{
        const int blocksize=1400;

	if (finished())
		return;

	if (local.data.fd==-1) {
		string basename=getfname();
		local.data.fd=::open(basename.c_str(), O_CREAT | O_RDWR | (local.data.truncate ? O_TRUNC : 0), 0644);
	}

        if ((ishdr->m_datalen != blocksize) && !(ishdr->isLastPacket())) {
            cerr << "Warning: Length of data " << ishdr->m_datalen << " != " << blocksize << endl;
        }
        __u8 *d=(__u8*)Data.getPointer(offset, ishdr->m_datalen);

	lseek(local.data.fd, blocksize * ishdr->getPktNr(), SEEK_SET);
	if (::write(local.data.fd, d, ishdr->m_datalen)!=ishdr->m_datalen) {
		cerr << "could not write all data" << endl;
	}
	setPacketDone(ishdr->getPktNr());

	if (local.data.bFinished && (local.data.fd!=-1))
		close();
}

daydata_BIN* daydata_BIN::allocate(PersistentStatus *status, InfosatHdr *ishdr)
{
	int packets=ishdr->getPktCount();

	int need_size=sizeof(daydata_BIN)+sizeof(__u8)*packets;

	daydata_BIN *obj=(daydata_BIN*) status->allocateBlock(need_size, ishdr->getMacID());
	if (obj!=NULL) {
		obj->init(ishdr);
	} else {
		cerr << "Did not get data-block" << endl;
	}
	return obj;
}

void daydata_BIN::printShort()
{
	cout << int(macID) << ":" << setw(2) << int(day) << "." << setw(2) << int(month);
}

int daydata_BIN::calcDayOffset()
{
	struct tm mytime;

	setenv("TZ","gmt",1);
	localtime_r(&programStartTime, &mytime);

	// Wenn gesendeter Tag Dezember und heute aber schon Januar
	if ((month==12) && (mytime.tm_mon=0)) {
		mytime.tm_year--;
	// und andersrum
	} else if ((month==1) && (mytime.tm_mon=11)) {
		mytime.tm_year++;
	}

	mytime.tm_mday=day;
	mytime.tm_mon=month-1;

	time_t t=mktime(&mytime);

	double diff=difftime(t,programStartTime);

	int fromtoday=int(diff/86400);

	int dayoffset=fromtoday-macID;
	return dayoffset;
}

void daydata_BIN::print()
{
	cout << int(macID) << ": " << setw(2) << int(day) << "." << setw(2) << int(month) << "  ";
	cout << setw(3) << goodPacketCount << "/" << setw(3) << packetsTotal << " ";

	if (verbose_level>1) {
		cout << "D" << setw(2) << getDayOffset() << " ";
	}

	if (finished()) {
		cout << "C  ";
	} else {
		cout << "   ";
	}
	char out='0';
	int good=0;
	int pkts=0;
	for (int i=0; i<packetsTotal; i++) {
		pkts++;
		if (goodPackets[i]) {
			good++;
		}
		out='0'+good;
		if (good==0)
			out='_';
		if (good==pkts)
			out='#';
		if (pkts==8) {
			cout << out;
			good=0;
			pkts=0;
		}
	}
	if (pkts)
		cout << out;
	cout << endl;
}

PersistentStatus::PersistentStatus(void)
{
	valid=false;
	truncate_output=false;
	history_macid=0;
	if (!openStatus())
		return;
	valid=true;
	ilog = new ofstream("./infosat.log");
	wakeup();
}

void PersistentStatus::wakeup(void)
{	
	goodPackets=0;
	dayoffset=-1000;
	for (int i=FIRST_DAY_NUMBER;i<=LAST_DAY_NUMBER;i++) {
		days[i]=(daydata_BIN*)getBlock(i);
		if (days[i]) {
			days[i]->wakeup();
			goodPackets += days[i]->goodPacketCount;
			if (days[i]->local.data.dayoffset > dayoffset)
				dayoffset=days[i]->local.data.dayoffset;
		}
	}
}

bool PersistentStatus::openStatus(void)
{
        fname = "infosat_status.dat";
	int fd=open(fname.c_str(), O_RDWR);
	if (fd==-1) {
		fd=open(fname.c_str(), O_RDWR | O_CREAT, 0644);
		truncate_output=true;
		write(fd, &id, sizeof(id));
		__u32 freestart=sizeof(statusMapHeader_BIN);
		
		write(fd, &freestart, sizeof(freestart));
			
		lseek(fd, mapSize-1, SEEK_SET);
		__u8 n=0;
		write(fd, &n, 1);
	}
	
	if (fd==-1)
		return false;
	
	mapdata=(__u8*)mmap(0, mapSize, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
	close(fd);
	if (!mapdata)
		return false;
	head = (statusMapHeader_BIN*)mapdata;
	
	for (int i=0;i<sizeof(id);i++) {
		if (id[i]!=head->id[i]) {
			cout << "Status-File invalid" << endl;
			cout << "(If this is correct, delete it)" << endl;
			return false;
		}
	}
	return true;
}

PersistentStatus::~PersistentStatus(void)
{
	delete ilog;
	ilog=NULL;
	if (finished()) {
		if (verbose_level>2) {
		} else {
			unlink(fname.c_str());
		}
	}
	if (mapdata) {
		munmap(mapdata, mapSize);
		mapdata=0;
		valid=false;
	}
}

void PersistentStatus::print(void)
{
	if (history_macid!=0) {
		cout << endl;
		if (days[history_macid]) {
			cout << "Last Packet: ";
			days[history_macid]->printShort();
			cout << "  " << setw(3) << (history_pktnr+1) << "/" << setw(3) << days[history_macid]->packetsTotal << endl;
		}
		for (int i=FIRST_DAY_NUMBER; i<=LAST_DAY_NUMBER; i++) {
			if (days[i]) {
				days[i]->print();
			}
		}
	}
}

__u8 *PersistentStatus::getBlock(int blockNr)
{
	int pos=head->day_start[blockNr];
	if (pos==0) {
		return 0;
	} else {
		return mapdata+pos;
	}
}

__u8 *PersistentStatus::allocateBlock(int size, int blockNr)
{
	if ((head->free_byte+size) <= mapSize) {
		int start=head->free_byte;
		__u8* obj=mapdata+start;
		head->free_byte += size;
		head->day_start[blockNr]=start;
		return obj;
	} else {
		cerr << "Could not allocate " << size << " Bytes" << endl;
		return 0;
	}
}

void PersistentStatus::ResetByOffset(void)
{
	cerr << "New Day Offset => Resetting" << endl;
	for (int i=FIRST_DAY_NUMBER;i<=LAST_DAY_NUMBER;i++) {
		if (days[i] && (days[i]->getDayOffset() < dayoffset)) {
			ResetDay(i);
		}
		if (days[i] && (days[i]->getDayOffset() > dayoffset)) {
			if (verbose_level>1) {
				cerr << "day with too high dayoffset found" << endl;
			}
		}
	}
}

void PersistentStatus::ResetDay(int macID)
{
	cerr << "Resetting day " << int(macID) << endl;
	days[macID]->close();
	days[macID]=0;
	head->day_start[macID]=0;
}

void PersistentStatus::InitDay(InfosatHdr *ishdr)
{
	int macID=ishdr->getMacID();
	days[macID]=daydata_BIN::allocate(this, ishdr);
	if (!days[macID])
		return;

	int thisoffset=days[macID]->getDayOffset();

	if (dayoffset != thisoffset) {
		if (verbose_level>1) {
			cout << "OLD OFFSET: " << dayoffset << endl;
			cout << "NEW OFFSET: " << days[macID]->getDayOffset() << endl;
		}
		if (dayoffset==-1000) {
			dayoffset=thisoffset;
		} else if (days[macID]->getDayOffset()>dayoffset) {
			dayoffset=thisoffset;
			ResetByOffset();
		}
	}
}

void PersistentStatus::Add(InfosatHdr *ishdr, RawData &Data, int &offset)
{
	//cerr << ">>>";
	int macID=ishdr->getMacID();

	if (days[macID]) {
		if ((days[macID]->day == ishdr->getDay()) && (days[macID]->month == ishdr->getMonth())) {
			// same day
			if (days[macID]->packetsTotal == ishdr->getPktCount()) {
				// GOOD, same Packetcount
			} else {
				// Packetcount changed, but same day
				ResetDay(macID);
			}
		} else {
			ResetDay(macID);
			InitDay(ishdr);
		}
	}

	if (!days[macID])
		InitDay(ishdr);

	if (!days[macID])
		return;

	if (days[macID]->finished()) {
		bool special_exit=true;
		for (int i=FIRST_DAY_NUMBER;i<=LAST_DAY_NUMBER; i++) {
			if (days[i]) {
				if (!days[i]->finished()) {
					special_exit=false;
				}
			}
		}
		if (special_exit) {
			cerr << "Special: All received days completed -> done" << endl;
			want_exit=true;
		}
	}

	days[macID]->write(ishdr, Data, offset);

	if (finished()) {
		want_exit=1;
	}

	history_macid=ishdr->getMacID();
	history_pktnr=ishdr->getPktNr();
	//cerr << "<<<  ";
}

bool PersistentStatus::finished()
{
	bool ret=true;
	for (int i=FIRST_DAY_NUMBER;i<=LAST_DAY_NUMBER;i++) {
		ret = ret && (days[i] && days[i]->finished());
	}
	return ret;
}

char *ipdump(iphdr *_ip, udphdr *_udp)
{
    static char dest[50];
    __u8 *addr;
    addr=(__u8*) (&(_ip->saddr));
    char sip[20];
    snprintf(sip,20, "%3d.%3d.%3d.%3d",addr[0],addr[1],addr[2],addr[3]);

    addr=(__u8*) (&(_ip->daddr));
    char dip[20];
    snprintf(dip,20, "%3d.%3d.%3d.%3d",addr[0],addr[1],addr[2],addr[3]);

    snprintf(dest,50,"[%s:%5d -> %s:%5d]", sip, _udp->source, dip, _udp->dest);
    return dest;
}

char *macdump(__u8 *mac)
{
    static char dest[20];
    snprintf(dest,20, "%02x:%02x:%02x:%02x:%02x:%02x",mac[0],mac[1],mac[2],mac[3],mac[4],mac[5]);
    return dest;
}

InfosatHdr::InfosatHdr(Section *_S)
{
    valid=false;
    s=_S;
    typ = NONE;
    ishdr = NULL;
}

InfosatHdr::~InfosatHdr(void)
{
}

void InfosatHdr::Parse(RawData Data, int &off)
{
    __u8 mac_mask[ETH_ALEN] = { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00 };
    __u8 mac[ETH_ALEN]      = { 0x01, 0x00, 0x5e, 0x02, 0x02, 0x00 };

    macID = s->m_ethhdr->h_dest[5];
    if (!s->testMACdest(mac_mask, mac))
        return;

    ishdr=(infosathdr_BIN*)Data.getPointer(off, sizeof(infosathdr_BIN));
    if (!ishdr)
        return;

    if (getPktNr() >= getPktCount())
        return;


    if ((macID>=FIRST_DAY_NUMBER) && (macID<=LAST_DAY_NUMBER)) {
        typ=EPG;
	if (ishdr->data[pos_day] > 31)
            return;
	if (ishdr->data[pos_month] > 12)
            return;
    }
    
    m_datalen = s->m_datalen - sizeof(infosathdr_BIN);
    off += sizeof(infosathdr_BIN);
    valid=true;
}

/*
__u32 InfosatHdr::getKey()
{
    return (macID<<16) | (TableID<<8);
}
*/
cInfosatReceiver::cInfosatReceiver()
{
    if (!epgstatus.valid) {
        exit(1);
    }
/*
    for (int i=0;i<DIMENSION_DAYS;i++) {
        days_done[i]=0;
        days_percent[i]=0;
    }
    m_wantAll=false;
    m_dEPGProgress=0.0;    
    m_info_day=0;
    m_info_month=0;
    m_info_macday=0;
*/
}


cInfosatReceiver::~cInfosatReceiver()
{
}


/*
void streamdump(UDPBody *udp, InfosatStream *stream)
{
    if (_s->m_haddumped)
        return;
    infosathdr *ishdr= (infosathdr*)(_s->getUDPDataStart());
    infosat_printf("%s %s ",macdump(_s->mac),ipdump(_s->m_ip,_s->m_udp));
    infosat_printf("#%2x  [%d/%d]", ishdr->TableID, ntohs(ishdr->pktnr)+1, ntohs(ishdr->pktcount));

    if (verbose_level>1) {
        infosat_printf("\n#%d:   ",ishdr->TableID);
        hexdump((__u8*)ishdr,sizeof(infosathdr));
    }
    
    _s->m_haddumped=true;
}

bool cInfosat::wantStream(InfosatHdr *ishdr)
{
    bool doit=false;

    switch(ishdr->TableID) {
    case TABLE_EPG:
        if ((ishdr->macID>=FIRST_DAY_NUMBER) && (ishdr->macID<=LAST_DAY_NUMBER))
            return true;
	break;
    default:
	return m_wantAll;
    }
    return false;
}
*/

void cInfosatReceiver::receiveSection(__u8 *_Data, int _Length)
{
    //cerr << "[";
    int offset=0;
    Section s(_Length);
    RawData rawdata(_Data, _Length);
    s.Parse(rawdata, offset);

    if ( !s.valid )
        return;

    if (s.typ != Section::UDP)
        return;

    InfosatHdr ishdr(&s);
    ishdr.Parse(rawdata, offset);

    if ( !ishdr.isValid() )
        return;

    switch (ishdr.getTyp()) {
        case InfosatHdr::EPG: {
            epgstatus.Add(&ishdr, rawdata, offset);
            break;
        }
    }

    
    /*
    InfosatStream *stream = streams[ishdr.getKey()];

    if (!stream) {
        if (wantStream(&ishdr)) {
            if (ishdr.isStartPacket()) {
                stream = InfosatStream::Decide(this, &ishdr);
                streams[ishdr.getKey()] = stream;
            } else {
                printf("%02d.%02d [ %d / %d ] ignored\n", ishdr.data[pos_day], ishdr.data[pos_month], ishdr.pktnr, ishdr.pktstotal);
                return;
            }
        } else {
            return;
        }
    }

    printf("%02d.%02d [ %3d - %3d ] good\n", ishdr.data[pos_day], ishdr.data[pos_month], ishdr.pktnr, ishdr.pktstotal);

    if (!stream)
        return;

    if (! stream->wantPacket(&ishdr)) {
        delete stream;
        stream=NULL;
        streams[ishdr.getKey()]=stream;
        return;
    }

    // Good packet received and want receive stream

    stream->AddData(&ishdr, rawdata, offset);
    if (stream->isFinished()) {
        delete stream;
        stream=NULL;
        streams[ishdr.getKey()]=stream;
    }
    printf("%2.1f%% done\n", m_dEPGProgress*100);
    */
}


void cInfosatReceiver::printStatus(void)
{
    epgstatus.print();
}




/*
void cInfosat::setEPGDayProgress(int macID, double pktnr, double pktcount)
{
    static int weights[LAST_DAY_NUMBER+1]={0,0x195,0x171,0x136,0x66,0x66,0x66,0x66};

    days_percent[macID]=pktnr/pktcount;    

    double p=0.0;
    double wsum=0.0;
    for (int day=1;day < LAST_DAY_NUMBER+1; day++) {
        wsum += weights[day];
        p += (weights[day]*days_percent[day]);
    }
    m_dEPGProgress = p / wsum;
}

void cInfosat::setEPGDayFinished(int macID)
{
    days_percent[macID]=1.0;
    days_done[macID]++;

    printf("DONE [");
    int i;
    for (i=FIRST_DAY_NUMBER;i<=LAST_DAY_NUMBER;i++) {
        infosat_printf(days_done[i] ? "X" : "-");
    }
    infosat_printf("]");
    for (i=FIRST_DAY_NUMBER;i<=LAST_DAY_NUMBER;i++) {
        if (days_done[i]==0) {
            return;
        }
    }
    want_exit=true;
}

void cInfosat::setEPGDayAborted(int macID)
{
    days_percent[macID]=0.0;
}

InfosatStream::InfosatStream(cInfosat *parent, InfosatHdr *ishdr)
{
    m_parent = parent;
    m_pktnr = -1;
    m_goodpkts = 0;
    m_bFinished = false;
    
    m_pktstotal = ishdr->pktstotal;
    m_fd=-1;
    m_TableID=ishdr->TableID;
    m_TableSubID=ishdr->TableSubID;

    m_basename = string("is_stream_") + int2string(ishdr->TableID, true);
    printf("NEW STREAM\n");
}

void InfosatStream::Open()
{
    if(m_fd!=-1)
        return;
    m_tmpfname = "tmp_" + m_basename;

    m_fd = open(m_tmpfname.c_str(), O_TRUNC | O_CREAT | O_RDWR, 0644);
    infosat_printf("writeing %s [Pkts: %d]\n", m_basename.c_str(), m_pktstotal);
    
}

void InfosatStream::Close()
{
    if(m_fd==-1)
        return;
    close(m_fd);
    string fname;
    if (m_bFinished) {
        fname = m_basename;
        infosat_printf("   FINISHED\n");
    } else {
        fname = "aborted_" + m_basename;
        infosat_printf("   aborted\n");
    }    
    rename(m_tmpfname.c_str(), fname.c_str());
}


InfosatStream::~InfosatStream()
{
    Close();
}

bool InfosatStream::wantPacket(InfosatHdr *ishdr)
{
    if ((m_pktnr+1) != ishdr->pktnr) {
        if (ishdr->pktnr>m_pktnr) {
            cerr << "Missed " << (ishdr->pktnr-m_pktnr+1) << " packets, aborting" << endl;
        } else {
            cerr << "Missed end of stream, aborting" << endl;
        }
        return false;
    }
    return true;
}

void InfosatStream::AddData(InfosatHdr *ishdr, RawData Data, int &off)
{
    __u8 *d=(__u8*)Data.getPointer(off, ishdr->datalength);
    if (!d) {
        printf("Could not get data!\n");
        return;
    }
    write(m_fd, d, ishdr->datalength);

    m_pktnr = ishdr->pktnr;

    UpdateStatus();
}

void InfosatStream::UpdateStatus()
{
    if (m_pktnr == (m_pktstotal-1)) {
        m_bFinished=true;
    }
}

void InfosatDayStream::UpdateStatus()
{
    InfosatStream::UpdateStatus();
    m_parent->setEPGDayProgress(macID, m_pktnr+1, m_pktstotal);
}

InfosatDayStream::InfosatDayStream(cInfosat *parent, InfosatHdr *ishdr):
    InfosatStream(parent, ishdr)
{
    day=ishdr->data[pos_day];
    month=ishdr->data[pos_month];
    macID = ishdr->macID;
    m_basename = string("infosat_") + int2string(month) + "_" + int2string(day);
}

InfosatDayStream::~InfosatDayStream()
{
    if (m_bFinished) {
        m_parent->setEPGDayFinished(macID);
    } else {
        m_parent->setEPGDayAborted(macID);
    }    
}

InfosatStream* InfosatStream::Decide(cInfosat *parent, InfosatHdr *ishdr)
{
    InfosatStream *s=NULL;
    switch (ishdr->TableID) {
      case TABLE_EPG:
        s = new InfosatDayStream(parent, ishdr);
        break;
      default:
        s = new InfosatStream(parent, ishdr);
        break;
    }
    if (s)
        s->Open();
    return s;
}
*/
