//
// C++ Interface: %{MODULE}
//
// Description: 
//
//
// Author: %{AUTHOR} <%{EMAIL}>, (C) %{YEAR}
//
// Copyright: See COPYING file that comes with this distribution
//
//
#ifndef CINFOSAT_H
#define CINFOSAT_H

#include "defaults.h"
#include "sectionreceiver.h"
#include <map>
#include <string>
#include "util.h"

using namespace std;

class Section;

#define FIRST_DAY_NUMBER 1
#define LAST_DAY_NUMBER 7
#define DIMENSION_DAYS (LAST_DAY_NUMBER+1)

/**
@author Matthias Schwarzott
*/

#define pos_day 0
#define pos_month 1


struct infosathdr_BIN {
	__u16	TechnisatID;
	__u8	TableID;
	__u8	TableSubID;
	__u8    data[4];
	__u16	pktnr;
	__u16	pktcount;
};

class PersistentStatus;
class InfosatHdr;

struct daydata_BIN {
public:
	// Header start
	__u8 macID;
	__u16 packetsTotal;
	__u16 goodPacketCount;
	__u8 day;
	__u8 month;
	union {
		__u8 space[20];
		struct {
			bool bFinished;
			bool bLogged;
			int fd;
			bool truncate;
			int dayoffset;
		} data;
	} local;
	// Header end
	__u8 goodPackets[0];


	static int mismatched;
	static time_t programStartTime;
	
	void setPacketDone(int i);
	bool finished() { return local.data.bFinished; }
	static daydata_BIN* allocate(PersistentStatus *status, InfosatHdr *ishdr);
	void init(InfosatHdr *ishdr);
	void wakeup(bool truncate=false);

	string getfname(void);
	void close();
	void log();
	void write(InfosatHdr *ishdr, RawData Data, int& offset);
	void print(void);
	void printShort(void);
	int calcDayOffset();
	int getDayOffset() { return local.data.dayoffset; };
};

class InfosatHdr : public Parsable {
private:
    Section *s;
    bool valid;

    infosathdr_BIN *ishdr;


    __u8 macID;
    
public:
    enum Type { NONE, EPG, UNKNOWN } typ;

    int m_datalen;

    InfosatHdr(Section *_S);
    ~InfosatHdr();
    virtual void Parse(RawData Data, int &off);
    virtual bool isValid() { return valid; }
    __u32 getKey();


    __u16 getPktNr() { return ntohs(ishdr->pktnr); }
    __u16 getPktCount() { return ntohs(ishdr->pktcount); }
    __u8 getDay() { return ishdr->data[pos_day]; }
    __u8 getMonth() { return ishdr->data[pos_month]; }

    bool isStartPacket() { return (getPktNr() == 0); }
    bool isLastPacket() { return (getPktNr() == (getPktCount()-1)); }

    __u8 getMacID() { return macID; }
    Type getTyp() { return typ; }
};

struct statusMapHeader_BIN {
	char id[12];
	__u32 free_byte;
	__u32 day_start[DIMENSION_DAYS];
};

class PersistentStatus {
private:
	string fname;
	__u8 *mapdata;
	statusMapHeader_BIN *head;


	daydata_BIN* days[DIMENSION_DAYS];
	static const int mapSize;
	static const char id[12];
	
	int goodPackets;
	
	__u8 history_macid;
	int history_pktnr;

	int dayoffset;
public:
	bool valid;
	bool truncate_output;
	PersistentStatus(void);
	~PersistentStatus(void);

	void wakeup(void);
	
	void Add(InfosatHdr *ishdr, RawData &Data, int &offset);
	bool finished(void);
	void print(void);

	void ResetByOffset(void);
	void ResetDay(int macID);
	void InitDay(InfosatHdr *ishdr);
	
	__u8* allocateBlock(int size, int blockNr);
	__u8* getBlock(int blockNr);
protected:
	bool openStatus();
};

class InfosatHdr;

class cInfosatReceiver : public cSectionReceiver {
private:
/*
    std::map<__u32, InfosatStream*> streams;

    char days_done[DIMENSION_DAYS];
    double days_percent[DIMENSION_DAYS];
    bool m_wantAll;
    
    double m_dEPGProgress;
    int m_info_day;
    int m_info_month;
    int m_info_macday;
*/
    
    PersistentStatus epgstatus;
public:
    cInfosatReceiver();
    ~cInfosatReceiver();
    void printStatus(void);
    virtual void receiveSection(__u8 *_Data, int _Length);


//    bool check_packet(infosathdr *hdr);
//    void stream_ended(cInfosatStream *stream, infosathdr *_hdr);

/*
    bool wantStream(InfosatHdr *ishdr);
    
    void wantAll(bool _wantAll) { m_wantAll=_wantAll; };
//    bool ProcessPacket(__u8 *mac, infosathdr *ishdr);
    
    double getEPGProgress(void) { return m_dEPGProgress; }

    void setEPGDayFinished(int macID);
    void setEPGDayAborted(int macID);
    void setEPGDayProgress(int macID, double pktnr, double pktcount);
//    infosatepghdr *getEPGhdr(infosathdr *_hdr) { return (infosatepghdr*) (&(_hdr->data)); }
*/
};


/*
class InfosatStream {
protected:
    int m_pktnr;
    int m_pktstotal;
    int m_goodpkts;

    bool m_bFinished;
    string m_basename;
    string m_tmpfname;
    int m_fd;
    
    int m_TableID;
    int m_TableSubID;
    cInfosat *m_parent;

public:
    InfosatStream(cInfosat *parent, InfosatHdr *ishdr);
    virtual ~InfosatStream();
    const string &getBasename() { return m_basename; };
    void Open();
    void Close();
    
    bool isFinished(void) { return m_bFinished; }

    static InfosatStream* Decide(cInfosat *parent, InfosatHdr *ishdr);
    virtual bool wantPacket(InfosatHdr *ishdr);
    void AddData(InfosatHdr *ishdr, RawData Data, int &off);
    virtual void UpdateStatus(void);
};

class InfosatDayStream : public InfosatStream {
public:
    __u8 day,month, macID;
    InfosatDayStream(cInfosat *parent, InfosatHdr *ishdr);
    virtual ~InfosatDayStream();
    virtual void UpdateStatus(void);
};
*/

#endif
