//
// C++ Implementation: %{MODULE}
//
// Description:
//
//
// Author: %{AUTHOR} <%{EMAIL}>, (C) %{YEAR}
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include "defaults.h"

#include <iostream>

#include "hexdump.h"
#include "util.h"
#include "section.h"
#include "checksums.h"
// #include "crc32.h"

#define SECTION_HEADER_SIZE 12
#define IP_HEADER_SIZE sizeof(iphdr)
#define UDP_HEADER_SIZE sizeof(udphdr)
#define SECTION_CRC_SIZE 4

#define IP_PROTOCOL_UDP 17
using namespace std;

Section::Section(int length) {
   valid=false;
   Length=length;

   m_iphdr=NULL;
   m_udphdr=NULL;
   typ = NONE;
}

Section::~Section()
{
}

void Section::Parse(RawData Data, int &off) {
   int startoffset = off;

   int crc_offset = Length - 4 - 1 + off;
   unsigned short section_len;

   section_len = ((Data[off+1] & 0xF) << 8) | (Data[off+2]) ;
   section_len += 3;/* 3 bytes before length count starts */

   if(section_len>Length)
       return;
   Length=section_len;

   //skip to end, read last 4 bytes and store in crc
   crc = Data.FourBytes(crc_offset);

   unsigned char *crc_ptr=Data.getPointer(off, section_len);

   table_id = Data[off++];
   flags[0] = Data[off++];
   flags[1] = Data[off++];



   eth.h_dest[5]=Data[off++];
   eth.h_dest[4]=Data[off++];

   flags2 = Data[off++];

   off += 2;

   eth.h_dest[3]=Data[off++];
   eth.h_dest[2]=Data[off++];
   eth.h_dest[1]=Data[off++];
   eth.h_dest[0]=Data[off++];

   m_ethhdr = &eth;
   for (int i=0;i<ETH_ALEN; i++) {
       eth.h_source[i]=0;
   }

   eth.h_proto=htons(ETH_P_IP);

   valid=true;

   switch (table_id) {
      case DSMCC_SECTION_MPE:
         // Test SNAP-Flag
         if ((flags2 & 2) == 0) {
            ParseIP(Data, off);
         }
   }
   m_datalen = Length - (off-startoffset) - 4;
}

bool Section::testMAC(__u8 mac[], __u8 mask[], __u8 mac_test[])
{
    bool good=true;
    for (int i=0; i<ETH_ALEN; i++) {
        if ((mac[i] & mask[i]) != mac_test[i])
            good=false;
    }
    return good;
}


bool Section::testMACdest(__u8 mask[], __u8 mac[])
{
    return testMAC(eth.h_dest, mask, mac);
}

void Section::ParseIP(RawData Data, int &off)
{
   m_iphdr=(iphdr*)Data.getPointer(off, sizeof(iphdr));
   if (!m_iphdr)
      return;
   if (m_iphdr->version != 4)
      return;

   if (IPChecksum(m_iphdr) != 0)
      return;

   off += sizeof(iphdr);
   typ = IP;

   switch (m_iphdr->protocol) {
      case 0x11:
         ParseUDP(Data, off);
         break;
   }
   return;
}

void Section::ParseUDP(RawData Data, int &off)
{
   m_udphdr=(udphdr*)Data.getPointer(off,sizeof(udphdr));

   int len_from_ip=ntohs(m_iphdr->tot_len);
   int len_from_udp=ntohs(m_udphdr->len);
   if (len_from_ip != len_from_udp+sizeof(iphdr))
   	return;

   if (UDPChecksum(m_iphdr, m_udphdr) != 0) {
       return;
   }

   typ=UDP;
   off += sizeof(udphdr);
}
