#include "defaults.h"

#include <stdio.h>
#include <stdlib.h>

#include <iostream>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "cdemux.h"
#include "cdemuxfile.h"
#include "cdemuxdevice.h"
#include "infosat/cinfosat.h"
#include "csectiondump.h"

using namespace std;

bool want_exit=false;
const char *default_output_dir="./received_data";
const char *output_dir=default_output_dir;
int verbose_level=1;
int SysLogLevel=3;

int main(int argc, char **argv)
{
	printf("infosatEPG\n");
	int adapter=0;
	int demux=0;
	int pid=1809;
	//pid=0x045a;
	char *channel=NULL;
	char *output_file=NULL;
	char *input_file=NULL;
	bool complete_stream=false;

	char opt;
	
	while ((opt = getopt(argc, argv, "ha:d:c:o:O:i:qvC")) != -1) {
		switch (opt) {
		case 'a':
			adapter = atoi(optarg);
			break;
		case 'd':
			demux = atoi(optarg);
			break;
		case 'c':
			channel = optarg;
			break;
		case 'o':
			output_dir = optarg;
			break;
		case 'O':
			output_file = optarg;
			break;
		case 'i':
			input_file = optarg;
			break;
		case 'q':
			verbose_level--;
			break;
		case 'v':
			verbose_level++;
			break;
		case 'C':
			complete_stream=true;
			break;
		case 'h':
			printf(	"Usage:\n"
				"	-a #     Number of dvb card to use (0,1,...)\n"
				"	-d #     Number of demux device to use (normally 0)\n"
				"	-c #     Channel to switch vdr to (example viva)\n"
				"	-o dir   Write output-files to dir (default: %s)\n"
				"	-h       Display this help\n"
				"	-O file  dumps the sections to file\n"
				"	-i file  read sections from file\n"
				"	-q       More quiet output\n"
				"	-v       More verbose output\n"
				"	-C       receive complete Stream\n",
				default_output_dir);
			exit(0);
			break;	
		}
	}

	struct stat stat_buf;

	cDemux *demuxer;
	if (input_file) {
		demuxer=new cDemuxFile(input_file);
	} else {
		if (channel) {
			char cmd[100];
			snprintf(cmd,100,"svdrpsend.pl chan %s",channel);
			system(cmd);
			sleep(1);
		}
		demuxer=new cDemuxDevice(adapter, demux, pid);
	}

	if (stat(output_dir, &stat_buf)!=0) {
		perror("output directory");
		exit(1);
	} else {
		chdir(output_dir);
	}

	
	cInfosatReceiver infosat;
	cSectionDump *dumper=NULL;

	if (output_file) {
		dumper=new cSectionDump(output_file);
		demuxer->Attach(dumper);
	}

	demuxer->Attach(&infosat);
	demuxer->Start();

	while(!want_exit) {
		infosat.printStatus();
		sleep(2);
	}
	infosat.printStatus();
	cout << "Finished reception" << endl;
}
